@@ -42,23 +42,23 @@
 #include <string.h>
 #include <stdio.h>
 
-#ifdef HAVE_SYS_SOCKET_H
+#if HAVE_SYS_SOCKET_H
 #include <sys/socket.h>
 #endif
 #include <sys/types.h>
-#ifdef HAVE_NETINET_IN_H
+#if HAVE_NETINET_IN_H
 #include <netinet/in.h>
 #endif
-#ifdef  HAVE_NETINET_IP_H
+#include <unistd.h>
+#if HAVE_NETINET_IP_H
 #include <netinet/ip.h>
 #endif
-#ifdef HAVE_NETINET_TCP_H
+#if HAVE_NETINET_TCP_H
 #include <netinet/tcp.h>
 #endif
 #include <sys/stat.h>
 #include <fcntl.h>
 
-#include "gstcurlelements.h"
 #include "gstcurlbasesink.h"
 #include "gstcurlfilesink.h"
 
@@ -94,8 +94,6 @@ static gboolean gst_curl_file_sink_prepare_transfer (GstCurlBaseSink *
 
 #define gst_curl_file_sink_parent_class parent_class
 G_DEFINE_TYPE (GstCurlFileSink, gst_curl_file_sink, GST_TYPE_CURL_BASE_SINK);
-GST_ELEMENT_REGISTER_DEFINE_WITH_CODE (curlfilesink, "curlfilesink",
-    GST_RANK_NONE, GST_TYPE_CURL_FILE_SINK, curl_element_init (plugin));
 
 static void
 gst_curl_file_sink_class_init (GstCurlFileSinkClass * klass)
@@ -106,6 +104,7 @@ gst_curl_file_sink_class_init (GstCurlFileSinkClass * klass)
 
   GST_DEBUG_CATEGORY_INIT (gst_curl_file_sink_debug, "curlfilesink", 0,
       "curl file sink element");
+  GST_DEBUG_OBJECT (klass, "class_init");
 
   gst_element_class_set_static_metadata (element_class,
       "Curl file sink",
@@ -234,7 +233,7 @@ gst_curl_file_sink_prepare_transfer (GstCurlBaseSink * basesink)
     if (last_slash != NULL) {
       /* create dir if file name contains dir component */
       gchar *dir_name = g_strndup (file_name, last_slash - file_name);
-      if (g_mkdir_with_parents (dir_name, 0755) < 0) {
+      if (g_mkdir_with_parents (dir_name, S_IRWXU) < 0) {
         basesink->error = g_strdup_printf ("failed to create directory '%s'",
             dir_name);
         g_free (file_name);
